@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2017 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2019 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -16,6 +16,7 @@
 
 #include "color.h"
 #include "bftw.h"
+#include "posix1e.h"
 #include "stat.h"
 #include "util.h"
 #include <assert.h>
@@ -28,6 +29,9 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
+/**
+ * A color for a file extension.
+ */
 struct ext_color {
 	const char *ext;
 	size_t len;
@@ -37,6 +41,9 @@ struct ext_color {
 	struct ext_color *next;
 };
 
+/**
+ * The parsed form of LS_COLORS.
+ */
 struct colors {
 	const char *reset;
 	const char *bold;
@@ -76,6 +83,9 @@ struct colors {
 	char *data;
 };
 
+/**
+ * A named color for parsing and lookup.
+ */
 struct color_name {
 	const char *name;
 	size_t offset;
@@ -83,6 +93,9 @@ struct color_name {
 
 #define COLOR_NAME(name, field) {name, offsetof(struct colors, field)}
 
+/**
+ * All the known color names that can be referenced in LS_COLORS.
+ */
 static const struct color_name color_names[] = {
 	COLOR_NAME("bd", block),
 	COLOR_NAME("bld", bold),
@@ -117,7 +130,8 @@ static const struct color_name color_names[] = {
 	{0},
 };
 
-static const char **look_up_color(const struct colors *colors, const char *name) {
+/** Get a color from the table. */
+static const char **get_color(const struct colors *colors, const char *name) {
 	for (const struct color_name *entry = color_names; entry->name; ++entry) {
 		if (strcmp(name, entry->name) == 0) {
 			return (const char **)((char *)colors + entry->offset);
@@ -127,17 +141,9 @@ static const char **look_up_color(const struct colors *colors, const char *name)
 	return NULL;
 }
 
-static const char *get_color(const struct colors *colors, const char *name) {
-	const char **color = look_up_color(colors, name);
-	if (color) {
-		return *color;
-	} else {
-		return NULL;
-	}
-}
-
+/** Set the value of a color. */
 static void set_color(struct colors *colors, const char *name, const char *value) {
-	const char **color = look_up_color(colors, name);
+	const char **color = get_color(colors, name);
 	if (color) {
 		*color = value;
 	}
@@ -192,28 +198,23 @@ struct colors *parse_colors(const char *ls_colors) {
 		goto done;
 	}
 
-	char *start = colors->data;
-	char *end;
-	struct ext_color *ext;
-	for (end = strchr(start, ':'); *start && end; start = end + 1, end = strchr(start, ':')) {
-		char *equals = strchr(start, '=');
+	for (char *chunk = colors->data, *next; chunk; chunk = next) {
+		next = strchr(chunk, ':');
+		if (next) {
+			*next++ = '\0';
+		}
+
+		char *equals = strchr(chunk, '=');
 		if (!equals) {
 			continue;
 		}
-
 		*equals = '\0';
-		*end = '\0';
 
-		const char *key = start;
+		const char *key = chunk;
 		const char *value = equals + 1;
 
-		// Ignore all-zero values
-		if (strspn(value, "0") == strlen(value)) {
-			continue;
-		}
-
 		if (key[0] == '*') {
-			ext = malloc(sizeof(struct ext_color));
+			struct ext_color *ext = malloc(sizeof(*ext));
 			if (ext) {
 				ext->ext = key + 1;
 				ext->len = strlen(ext->ext);
@@ -222,6 +223,11 @@ struct colors *parse_colors(const char *ls_colors) {
 				colors->ext_list = ext;
 			}
 		} else {
+			// All-zero values should be treated like NULL, to fall
+			// back on any other relevant coloring for that file
+			if (strcmp(key, "rs") != 0 && strspn(value, "0") == strlen(value)) {
+				value = NULL;
+			}
 			set_color(colors, key, value);
 		}
 	}
@@ -296,10 +302,33 @@ int cfclose(CFILE *cfile) {
 	return ret;
 }
 
+/** Get the color for a file by its extension. */
+static const char *ext_color(const struct colors *colors, const char *filename) {
+	size_t namelen = strlen(filename);
+
+	for (const struct ext_color *ext = colors->ext_list; ext; ext = ext->next) {
+		if (namelen < ext->len) {
+			continue;
+		}
+
+		const char *suffix = filename + namelen - ext->len;
+		if (strcasecmp(suffix, ext->ext) == 0) {
+			return ext->color;
+		}
+	}
+
+	return NULL;
+}
+
+/** Get the color for a file. */
 static const char *file_color(const struct colors *colors, const char *filename, const struct BFTW *ftwbuf) {
 	const struct bfs_stat *sb = ftwbuf->statbuf;
 	if (!sb) {
-		return colors->orphan;
+		if (colors->missing) {
+			return colors->missing;
+		} else {
+			return colors->orphan;
+		}
 	}
 
 	const char *color = NULL;
@@ -310,6 +339,8 @@ static const char *file_color(const struct colors *colors, const char *filename,
 			color = colors->setuid;
 		} else if (sb->mode & S_ISGID) {
 			color = colors->setgid;
+		} else if (colors->capable && bfs_check_capabilities(ftwbuf)) {
+			color = colors->capable;
 		} else if (sb->mode & 0111) {
 			color = colors->exec;
 		}
@@ -319,14 +350,7 @@ static const char *file_color(const struct colors *colors, const char *filename,
 		}
 
 		if (!color) {
-			size_t namelen = strlen(filename);
-
-			for (struct ext_color *ext = colors->ext_list; ext; ext = ext->next) {
-				if (namelen >= ext->len && memcmp(filename + namelen - ext->len, ext->ext, ext->len) == 0) {
-					color = ext->color;
-					break;
-				}
-			}
+			color = ext_color(colors, filename);
 		}
 
 		if (!color) {
@@ -353,9 +377,8 @@ static const char *file_color(const struct colors *colors, const char *filename,
 		break;
 
 	case S_IFLNK:
-		if (xfaccessat(ftwbuf->at_fd, ftwbuf->at_path, F_OK) == 0) {
-			color = colors->link;
-		} else {
+		color = colors->link;
+		if (colors->orphan && xfaccessat(ftwbuf->at_fd, ftwbuf->at_path, F_OK) != 0) {
 			color = colors->orphan;
 		}
 		break;
@@ -387,6 +410,7 @@ static const char *file_color(const struct colors *colors, const char *filename,
 	return color;
 }
 
+/** Print an ANSI escape sequence. */
 static int print_esc(const char *esc, FILE *file) {
 	if (fputs("\033[", file) == EOF) {
 		return -1;
@@ -401,6 +425,26 @@ static int print_esc(const char *esc, FILE *file) {
 	return 0;
 }
 
+/** Print a string with an optional color. */
+static int print_colored(const struct colors *colors, const char *esc, const char *str, size_t len, FILE *file) {
+	if (esc) {
+		if (print_esc(esc, file) != 0) {
+			return -1;
+		}
+	}
+	if (fwrite(str, 1, len, file) != len) {
+		return -1;
+	}
+	if (esc) {
+		if (print_esc(colors->reset, file) != 0) {
+			return -1;
+		}
+	}
+
+	return 0;
+}
+
+/** Print the path to a file with the appropriate colors. */
 static int print_path(CFILE *cfile, const struct BFTW *ftwbuf) {
 	const struct colors *colors = cfile->colors;
 	FILE *file = cfile->file;
@@ -410,40 +454,22 @@ static int print_path(CFILE *cfile, const struct BFTW *ftwbuf) {
 		return fputs(path, file) == EOF ? -1 : 0;
 	}
 
+	if (ftwbuf->nameoff > 0) {
+		if (print_colored(colors, colors->dir, path, ftwbuf->nameoff, file) != 0) {
+			return -1;
+		}
+	}
+
 	const char *filename = path + ftwbuf->nameoff;
-
-	if (colors->dir) {
-		if (print_esc(colors->dir, file) != 0) {
-			return -1;
-		}
-	}
-	if (fwrite(path, 1, ftwbuf->nameoff, file) != ftwbuf->nameoff) {
-		return -1;
-	}
-	if (colors->dir) {
-		if (print_esc(colors->reset, file) != 0) {
-			return -1;
-		}
-	}
-
 	const char *color = file_color(colors, filename, ftwbuf);
-	if (color) {
-		if (print_esc(color, file) != 0) {
-			return -1;
-		}
-	}
-	if (fputs(filename, file) == EOF) {
+	if (print_colored(colors, color, filename, strlen(filename), file) != 0) {
 		return -1;
 	}
-	if (color) {
-		if (print_esc(colors->reset, file) != 0) {
-			return -1;
-		}
-	}
 
 	return 0;
 }
 
+/** Print a link target with the appropriate colors. */
 static int print_link(CFILE *cfile, const struct BFTW *ftwbuf) {
 	int ret = -1;
 
@@ -471,128 +497,152 @@ done:
 }
 
 int cfprintf(CFILE *cfile, const char *format, ...) {
-	const struct colors *colors = cfile->colors;
-	FILE *file = cfile->file;
-
-	int ret = -1;
-	int error = errno;
-
 	va_list args;
 	va_start(args, format);
+	int ret = cvfprintf(cfile, format, args);
+	va_end(args);
+	return ret;
+}
+
+int cvfprintf(CFILE *cfile, const char *format, va_list args) {
+	const struct colors *colors = cfile->colors;
+	FILE *file = cfile->file;
+	int error = errno;
 
 	for (const char *i = format; *i; ++i) {
-		const char *percent = strchr(i, '%');
-		if (!percent) {
-			if (fputs(i, file) == EOF) {
-				goto done;
-			}
-			break;
+		size_t verbatim = strcspn(i, "%$");
+		if (fwrite(i, 1, verbatim, file) != verbatim) {
+			return -1;
 		}
 
-		size_t len = percent - i;
-		if (fwrite(i, 1, len, file) != len) {
-			goto done;
-		}
-
-		i = percent + 1;
+		i += verbatim;
 		switch (*i) {
 		case '%':
-			if (fputc('%', file) == EOF) {
-				goto done;
-			}
-			break;
+			switch (*++i) {
+			case '%':
+				if (fputc('%', file) == EOF) {
+					return -1;
+				}
+				break;
 
-		case 'c':
-			if (fputc(va_arg(args, int), file) == EOF) {
-				goto done;
-			}
-			break;
+			case 'c':
+				if (fputc(va_arg(args, int), file) == EOF) {
+					return -1;
+				}
+				break;
 
-		case 'd':
-			if (fprintf(file, "%d", va_arg(args, int)) < 0) {
-				goto done;
-			}
-			break;
+			case 'd':
+				if (fprintf(file, "%d", va_arg(args, int)) < 0) {
+					return -1;
+				}
+				break;
 
-		case 'g':
-			if (fprintf(file, "%g", va_arg(args, double)) < 0) {
-				goto done;
-			}
-			break;
+			case 'g':
+				if (fprintf(file, "%g", va_arg(args, double)) < 0) {
+					return -1;
+				}
+				break;
 
-		case 's':
-			if (fputs(va_arg(args, const char *), file) == EOF) {
-				goto done;
-			}
-			break;
+			case 's':
+				if (fputs(va_arg(args, const char *), file) == EOF) {
+					return -1;
+				}
+				break;
 
-		case 'z':
-			++i;
-			if (*i != 'u') {
+			case 'z':
+				++i;
+				if (*i != 'u') {
+					goto invalid;
+				}
+				if (fprintf(file, "%zu", va_arg(args, size_t)) < 0) {
+					return -1;
+				}
+				break;
+
+			case 'm':
+				if (fputs(strerror(error), file) == EOF) {
+					return -1;
+				}
+				break;
+
+			case 'p':
+				switch (*++i) {
+				case 'P':
+					if (print_path(cfile, va_arg(args, const struct BFTW *)) != 0) {
+						return -1;
+					}
+					break;
+
+				case 'L':
+					if (print_link(cfile, va_arg(args, const struct BFTW *)) != 0) {
+						return -1;
+					}
+					break;
+
+				default:
+					goto invalid;
+				}
+
+				break;
+
+			default:
 				goto invalid;
 			}
-			if (fprintf(file, "%zu", va_arg(args, size_t)) < 0) {
-				goto done;
-			}
 			break;
 
-		case 'm':
-			if (fputs(strerror(error), file) == EOF) {
-				goto done;
-			}
-			break;
+		case '$':
+			switch (*++i) {
+			case '$':
+				if (fputc('$', file) == EOF) {
+					return -1;
+				}
+				break;
 
-		case 'P':
-			if (print_path(cfile, va_arg(args, const struct BFTW *)) != 0) {
-				goto done;
-			}
-			break;
+			case '{': {
+				++i;
+				const char *end = strchr(i, '}');
+				if (!end) {
+					goto invalid;
+				}
+				if (!colors) {
+					i = end;
+					break;
+				}
 
-		case 'L':
-			if (print_link(cfile, va_arg(args, const struct BFTW *)) != 0) {
-				goto done;
-			}
-			break;
+				size_t len = end - i;
+				char name[len + 1];
+				memcpy(name, i, len);
+				name[len] = '\0';
+
+				const char **esc = get_color(colors, name);
+				if (!esc) {
+					goto invalid;
+				}
+				if (*esc) {
+					if (print_esc(*esc, file) != 0) {
+						return -1;
+					}
+				}
 
-		case '{': {
-			++i;
-			const char *end = strchr(i, '}');
-			if (!end) {
-				goto invalid;
-			}
-			if (!colors) {
 				i = end;
 				break;
 			}
 
-			size_t len = end - i;
-			char name[len + 1];
-			memcpy(name, i, len);
-			name[len] = '\0';
-
-			const char *esc = get_color(colors, name);
-			if (!esc) {
+			default:
 				goto invalid;
 			}
-			if (print_esc(esc, file) != 0) {
-				goto done;
-			}
-
-			i = end;
 			break;
-		}
 
 		default:
-		invalid:
-			assert(false);
-			errno = EINVAL;
-			goto done;
+			return 0;
 		}
+
 	}
 
-	ret = 0;
+	return 0;
 
-done:
-	va_end(args);
-	return ret;
+invalid:
+	assert(false);
+	errno = EINVAL;
+	return -1;
 }
