@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2017 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2018 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -14,13 +14,19 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
  ****************************************************************************/
 
+/**
+ * Implementation of all the literal expressions.
+ */
+
 #include "eval.h"
 #include "bftw.h"
 #include "cmdline.h"
 #include "color.h"
+#include "diag.h"
 #include "dstring.h"
 #include "exec.h"
 #include "mtab.h"
+#include "posix1e.h"
 #include "printf.h"
 #include "stat.h"
 #include "util.h"
@@ -31,6 +37,7 @@
 #include <fnmatch.h>
 #include <grp.h>
 #include <pwd.h>
+#include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -57,33 +64,16 @@ struct eval_state {
 	struct bfs_stat xstatbuf;
 };
 
-/**
- * Check if an error should be ignored.
- */
-static bool eval_should_ignore(const struct eval_state *state, int error) {
-	return state->cmdline->ignore_races
-		&& is_nonexistence_error(error)
-		&& state->ftwbuf->depth > 0;
-}
-
-/**
- * Report an error that occurs during evaluation.
- */
-static void eval_error(struct eval_state *state) {
-	if (!eval_should_ignore(state, errno)) {
-		cfprintf(state->cmdline->cerr, "%{er}error: '%s': %m%{rs}\n", state->ftwbuf->path);
-		*state->ret = EXIT_FAILURE;
-	}
-}
-
-#define DEBUG_FLAG(flags, flag)			\
-	if ((flags & flag) || flags == flag) {	\
-		fputs(#flag, stderr);		\
-		flags ^= flag;			\
-		if (flags) {			\
-			fputs(" | ", stderr);	\
-		}				\
-	}
+#define DEBUG_FLAG(flags, flag)				\
+	do {						\
+		if ((flags & flag) || flags == flag) {	\
+			fputs(#flag, stderr);		\
+			flags ^= flag;			\
+			if (flags) {			\
+				fputs(" | ", stderr);	\
+			}				\
+		}					\
+	} while (0)
 
 /**
  * Debug stat() calls.
@@ -121,19 +111,82 @@ static void debug_stat(const struct eval_state *state, int at_flags, enum bfs_st
 /**
  * Perform a bfs_stat() call if necessary.
  */
-static const struct bfs_stat *eval_stat(struct eval_state *state) {
+static const struct bfs_stat *eval_try_stat(struct eval_state *state) {
 	struct BFTW *ftwbuf = state->ftwbuf;
-	if (!ftwbuf->statbuf) {
-		debug_stat(state, ftwbuf->at_flags, BFS_STAT_BROKEN_OK);
-		if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, ftwbuf->at_flags, BFS_STAT_BROKEN_OK, &state->statbuf) == 0) {
-			ftwbuf->statbuf = &state->statbuf;
-		} else {
-			eval_error(state);
-		}
+
+	if (ftwbuf->statbuf) {
+		goto done;
 	}
+
+	if (ftwbuf->error) {
+		errno = ftwbuf->error;
+		goto done;
+	}
+
+	debug_stat(state, ftwbuf->at_flags, BFS_STAT_BROKEN_OK);
+
+	if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, ftwbuf->at_flags, BFS_STAT_BROKEN_OK, &state->statbuf) == 0) {
+		ftwbuf->statbuf = &state->statbuf;
+	} else {
+		ftwbuf->error = errno;
+	}
+
+done:
 	return ftwbuf->statbuf;
 }
 
+/**
+ * Print an error message.
+ */
+BFS_FORMATTER(2, 3)
+static void eval_error(struct eval_state *state, const char *format, ...) {
+	int error = errno;
+	const struct cmdline *cmdline = state->cmdline;
+	CFILE *cerr = cmdline->cerr;
+
+	if (cerr->colors) {
+		eval_try_stat(state);
+	}
+
+	bfs_error(cmdline, "%pP: ", state->ftwbuf);
+
+	va_list args;
+	va_start(args, format);
+	errno = error;
+	cvfprintf(cerr, format, args);
+	va_end(args);
+}
+
+/**
+ * Check if an error should be ignored.
+ */
+static bool eval_should_ignore(const struct eval_state *state, int error) {
+	return state->cmdline->ignore_races
+		&& is_nonexistence_error(error)
+		&& state->ftwbuf->depth > 0;
+}
+
+/**
+ * Report an error that occurs during evaluation.
+ */
+static void eval_report_error(struct eval_state *state) {
+	if (!eval_should_ignore(state, errno)) {
+		eval_error(state, "%m.\n");
+		*state->ret = EXIT_FAILURE;
+	}
+}
+
+/**
+ * Perform a bfs_stat() call if necessary.
+ */
+static const struct bfs_stat *eval_stat(struct eval_state *state) {
+	const struct bfs_stat *ret = eval_try_stat(state);
+	if (!ret) {
+		eval_report_error(state);
+	}
+	return ret;
+}
+
 /**
  * Perform a bfs_stat() call for tests that flip the follow flag, like -xtype.
  */
@@ -196,42 +249,42 @@ bool eval_access(const struct expr *expr, struct eval_state *state) {
 	return xfaccessat(ftwbuf->at_fd, ftwbuf->at_path, expr->idata) == 0;
 }
 
+/**
+ * -acl test.
+ */
+bool eval_acl(const struct expr *expr, struct eval_state *state) {
+	return bfs_check_acl(state->ftwbuf);
+}
+
+/**
+ * -capable test.
+ */
+bool eval_capable(const struct expr *expr, struct eval_state *state) {
+	return bfs_check_capabilities(state->ftwbuf);
+}
+
 /**
  * Get the given timespec field out of a stat buffer.
  */
-static const struct timespec *eval_stat_time(const struct expr *expr, struct eval_state *state) {
-	const struct bfs_stat *statbuf = eval_stat(state);
-	if (!statbuf) {
-		return NULL;
-	}
-
-	if (!(statbuf->mask & expr->stat_field)) {
-		assert(expr->stat_field == BFS_STAT_BTIME);
-		cfprintf(state->cmdline->cerr, "%{er}error: '%s': Couldn't get file birth time.%{rs}\n", state->ftwbuf->path);
+static const struct timespec *eval_stat_time(const struct bfs_stat *statbuf, enum bfs_stat_field field, struct eval_state *state) {
+	const struct timespec *ret = bfs_stat_time(statbuf, field);
+	if (!ret) {
+		eval_error(state, "Couldn't get file %s: %m.\n", bfs_stat_field_name(field));
 		*state->ret = EXIT_FAILURE;
-		return NULL;
-	}
-
-	switch (expr->stat_field) {
-	case BFS_STAT_ATIME:
-		return &statbuf->atime;
-	case BFS_STAT_BTIME:
-		return &statbuf->btime;
-	case BFS_STAT_CTIME:
-		return &statbuf->ctime;
-	case BFS_STAT_MTIME:
-		return &statbuf->mtime;
-	default:
-		assert(false);
-		return NULL;
 	}
+	return ret;
 }
 
 /**
  * -[aBcm]?newer tests.
  */
 bool eval_newer(const struct expr *expr, struct eval_state *state) {
-	const struct timespec *time = eval_stat_time(expr, state);
+	const struct bfs_stat *statbuf = eval_stat(state);
+	if (!statbuf) {
+		return false;
+	}
+
+	const struct timespec *time = eval_stat_time(statbuf, expr->stat_field, state);
 	if (!time) {
 		return false;
 	}
@@ -244,7 +297,12 @@ bool eval_newer(const struct expr *expr, struct eval_state *state) {
  * -[aBcm]{min,time} tests.
  */
 bool eval_time(const struct expr *expr, struct eval_state *state) {
-	const struct timespec *time = eval_stat_time(expr, state);
+	const struct bfs_stat *statbuf = eval_stat(state);
+	if (!statbuf) {
+		return false;
+	}
+
+	const struct timespec *time = eval_stat_time(statbuf, expr->stat_field, state);
 	if (!time) {
 		return false;
 	}
@@ -271,7 +329,13 @@ bool eval_used(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	time_t diff = timespec_diff(&statbuf->atime, &statbuf->ctime);
+	const struct timespec *atime = eval_stat_time(statbuf, BFS_STAT_ATIME, state);
+	const struct timespec *ctime = eval_stat_time(statbuf, BFS_STAT_CTIME, state);
+	if (!atime || !ctime) {
+		return false;
+	}
+
+	time_t diff = timespec_diff(atime, ctime);
 	diff /= 60*60*24;
 	return expr_cmp(expr, diff);
 }
@@ -309,7 +373,16 @@ bool eval_nogroup(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	return getgrgid(statbuf->gid) == NULL;
+	errno = 0;
+	if (getgrgid(statbuf->gid) == NULL) {
+		if (errno == 0) {
+			return true;
+		} else {
+			eval_report_error(state);
+		}
+	}
+
+	return false;
 }
 
 /**
@@ -321,7 +394,16 @@ bool eval_nouser(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	return getpwuid(statbuf->uid) == NULL;
+	errno = 0;
+	if (getpwuid(statbuf->uid) == NULL) {
+		if (errno == 0) {
+			return true;
+		} else {
+			eval_report_error(state);
+		}
+	}
+
+	return false;
 }
 
 /**
@@ -348,13 +430,13 @@ bool eval_delete(const struct expr *expr, struct eval_state *state) {
 				flag |= AT_REMOVEDIR;
 			}
 		} else {
-			eval_error(state);
+			eval_report_error(state);
 			return false;
 		}
 	}
 
 	if (unlinkat(ftwbuf->at_fd, ftwbuf->at_path, flag) != 0) {
-		eval_error(state);
+		eval_report_error(state);
 		return false;
 	}
 
@@ -366,7 +448,7 @@ static int eval_exec_finish(const struct expr *expr, const struct cmdline *cmdli
 	int ret = 0;
 	if (expr->execbuf && bfs_exec_finish(expr->execbuf) != 0) {
 		if (errno != 0) {
-			cfprintf(cmdline->cerr, "%{er}error: %s %s: %m%{rs}\n", expr->argv[0], expr->argv[1]);
+			bfs_error(cmdline, "%s %s: %m.\n", expr->argv[0], expr->argv[1]);
 		}
 		ret = -1;
 	}
@@ -385,7 +467,7 @@ static int eval_exec_finish(const struct expr *expr, const struct cmdline *cmdli
 bool eval_exec(const struct expr *expr, struct eval_state *state) {
 	bool ret = bfs_exec(expr->execbuf, state->ftwbuf) == 0;
 	if (errno != 0) {
-		cfprintf(state->cmdline->cerr, "%{er}error: %s %s: %m%{rs}\n", expr->argv[0], expr->argv[1]);
+		eval_error(state, "%s %s: %m.\n", expr->argv[0], expr->argv[1]);
 		*state->ret = EXIT_FAILURE;
 	}
 	return ret;
@@ -418,13 +500,13 @@ bool eval_empty(const struct expr *expr, struct eval_state *state) {
 	if (ftwbuf->typeflag == BFTW_DIR) {
 		int dfd = openat(ftwbuf->at_fd, ftwbuf->at_path, O_RDONLY | O_CLOEXEC | O_DIRECTORY);
 		if (dfd < 0) {
-			eval_error(state);
+			eval_report_error(state);
 			goto done;
 		}
 
 		DIR *dir = fdopendir(dfd);
 		if (!dir) {
-			eval_error(state);
+			eval_report_error(state);
 			close(dfd);
 			goto done;
 		}
@@ -434,7 +516,7 @@ bool eval_empty(const struct expr *expr, struct eval_state *state) {
 		while (true) {
 			struct dirent *de;
 			if (xreaddir(dir, &de) != 0) {
-				eval_error(state);
+				eval_report_error(state);
 				goto done_dir;
 			}
 			if (!de) {
@@ -536,7 +618,7 @@ bool eval_lname(const struct expr *expr, struct eval_state *state) {
 
 	name = xreadlinkat(ftwbuf->at_fd, ftwbuf->at_path, statbuf->size);
 	if (!name) {
-		eval_error(state);
+		eval_report_error(state);
 		goto done;
 	}
 
@@ -562,7 +644,7 @@ bool eval_name(const struct expr *expr, struct eval_state *state) {
 		if (slash && slash > name) {
 			copy = strndup(name, slash - name);
 			if (!copy) {
-				eval_error(state);
+				eval_report_error(state);
 				return false;
 			}
 			name = copy;
@@ -626,7 +708,7 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 	}
 
 	uintmax_t ino = statbuf->ino;
-	uintmax_t blocks = (statbuf->blocks + 1)/2;
+	uintmax_t blocks = ((uintmax_t)statbuf->blocks*BFS_STAT_BLKSIZE + 1023)/1024;
 	char mode[11];
 	format_mode(statbuf->mode, mode);
 	uintmax_t nlink = statbuf->nlink;
@@ -658,9 +740,17 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 		}
 	}
 
-	uintmax_t size = statbuf->size;
-	if (fprintf(file, " %8ju", size) < 0) {
-		goto error;
+	if (ftwbuf->typeflag & (BFTW_BLK | BFTW_CHR)) {
+		int ma = bfs_major(statbuf->rdev);
+		int mi = bfs_minor(statbuf->rdev);
+		if (fprintf(file, " %3d, %3d", ma, mi) < 0) {
+			goto error;
+		}
+	} else {
+		uintmax_t size = statbuf->size;
+		if (fprintf(file, " %8ju", size) < 0) {
+			goto error;
+		}
 	}
 
 	time_t time = statbuf->mtime.tv_sec;
@@ -684,12 +774,12 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 		goto error;
 	}
 
-	if (cfprintf(cfile, " %P", ftwbuf) < 0) {
+	if (cfprintf(cfile, " %pP", ftwbuf) < 0) {
 		goto error;
 	}
 
 	if (ftwbuf->typeflag == BFTW_LNK) {
-		if (cfprintf(cfile, " -> %L", ftwbuf) < 0) {
+		if (cfprintf(cfile, " -> %pL", ftwbuf) < 0) {
 			goto error;
 		}
 	}
@@ -702,7 +792,7 @@ done:
 	return true;
 
 error:
-	eval_error(state);
+	eval_report_error(state);
 	return true;
 }
 
@@ -715,8 +805,8 @@ bool eval_fprint(const struct expr *expr, struct eval_state *state) {
 		eval_stat(state);
 	}
 
-	if (cfprintf(cfile, "%P\n", state->ftwbuf) < 0) {
-		eval_error(state);
+	if (cfprintf(cfile, "%pP\n", state->ftwbuf) < 0) {
+		eval_report_error(state);
 	}
 
 	return true;
@@ -729,7 +819,7 @@ bool eval_fprint0(const struct expr *expr, struct eval_state *state) {
 	const char *path = state->ftwbuf->path;
 	size_t length = strlen(path) + 1;
 	if (fwrite(path, 1, length, expr->cfile->file) != length) {
-		eval_error(state);
+		eval_report_error(state);
 	}
 	return true;
 }
@@ -745,7 +835,7 @@ bool eval_fprintf(const struct expr *expr, struct eval_state *state) {
 	}
 
 	if (bfs_printf(expr->cfile->file, expr->printf, state->ftwbuf) != 0) {
-		eval_error(state);
+		eval_report_error(state);
 	}
 
 done:
@@ -786,7 +876,7 @@ bool eval_fprintx(const struct expr *expr, struct eval_state *state) {
 	return true;
 
 error:
-	eval_error(state);
+	eval_report_error(state);
 	return true;
 }
 
@@ -828,7 +918,7 @@ bool eval_regex(const struct expr *expr, struct eval_state *state) {
 	} else if (err != REG_NOMATCH) {
 		char *str = xregerror(err, expr->regex);
 		if (str) {
-			cfprintf(state->cmdline->cerr, "%{er}error: '%s': %s%{rs}\n", path, str);
+			eval_error(state, "%s.\n", str);
 			free(str);
 		} else {
 			perror("xregerror()");
@@ -886,7 +976,7 @@ bool eval_sparse(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	blkcnt_t expected = (statbuf->size + 511)/512;
+	blkcnt_t expected = (statbuf->size + BFS_STAT_BLKSIZE - 1)/BFS_STAT_BLKSIZE;
 	return statbuf->blocks < expected;
 }
 
@@ -911,12 +1001,12 @@ bool eval_xtype(const struct expr *expr, struct eval_state *state) {
 
 	const struct bfs_stat *statbuf = eval_xstat(state);
 	if (statbuf) {
-		return mode_to_typeflag(statbuf->mode) & expr->idata;
+		return bftw_mode_typeflag(statbuf->mode) & expr->idata;
 	} else if (!follow && is_nonexistence_error(errno)) {
 		// Broken symlink
 		return eval_type(expr, state);
 	} else {
-		eval_error(state);
+		eval_report_error(state);
 		return false;
 	}
 }
@@ -1132,7 +1222,7 @@ static enum bftw_action cmdline_callback(struct BFTW *ftwbuf, void *ptr) {
 	if (ftwbuf->typeflag == BFTW_ERROR) {
 		if (!eval_should_ignore(&state, ftwbuf->error)) {
 			args->ret = EXIT_FAILURE;
-			cfprintf(cmdline->cerr, "%{er}error: '%s': %s%{rs}\n", ftwbuf->path, strerror(ftwbuf->error));
+			eval_error(&state, "%s.\n", strerror(ftwbuf->error));
 		}
 		state.action = BFTW_SKIP_SUBTREE;
 		goto done;
@@ -1140,7 +1230,7 @@ static enum bftw_action cmdline_callback(struct BFTW *ftwbuf, void *ptr) {
 
 	if (cmdline->xargs_safe && strpbrk(ftwbuf->path, " \t\n\'\"\\")) {
 		args->ret = EXIT_FAILURE;
-		cfprintf(cmdline->cerr, "%{er}error: '%s': Path is not safe for xargs.%{rs}\n", ftwbuf->path);
+		eval_error(&state, "Path is not safe for xargs.\n");
 		state.action = BFTW_SKIP_SUBTREE;
 		goto done;
 	}
