@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2017 Tavian Barnes <tavianator@tavianator.com>             *
+ * Copyright (C) 2017-2018 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -17,6 +17,7 @@
 #include "printf.h"
 #include "cmdline.h"
 #include "color.h"
+#include "diag.h"
 #include "dstring.h"
 #include "expr.h"
 #include "mtab.h"
@@ -76,37 +77,13 @@ static int bfs_printf_flush(FILE *file, const struct bfs_printf_directive *direc
 	assert(ret >= 0 && ret < sizeof(buf));				\
 	(void)ret
 
-/**
- * Get a particular time field from a struct bfs_stat.
- */
-static const struct timespec *get_time_field(const struct bfs_stat *statbuf, enum bfs_stat_field stat_field) {
-	if (!(statbuf->mask & stat_field)) {
-		errno = ENOTSUP;
-		return NULL;
-	}
-
-	switch (stat_field) {
-	case BFS_STAT_ATIME:
-		return &statbuf->atime;
-	case BFS_STAT_BTIME:
-		return &statbuf->btime;
-	case BFS_STAT_CTIME:
-		return &statbuf->ctime;
-	case BFS_STAT_MTIME:
-		return &statbuf->mtime;
-	default:
-		assert(false);
-		return NULL;
-	}
-}
-
-/** %c, %c, and %t: ctime() */
+/** %a, %c, %t: ctime() */
 static int bfs_printf_ctime(FILE *file, const struct bfs_printf_directive *directive, const struct BFTW *ftwbuf) {
 	// Not using ctime() itself because GNU find adds nanoseconds
 	static const char *days[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
 	static const char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
 
-	const struct timespec *ts = get_time_field(ftwbuf->statbuf, directive->stat_field);
+	const struct timespec *ts = bfs_stat_time(ftwbuf->statbuf, directive->stat_field);
 	if (!ts) {
 		return -1;
 	}
@@ -129,9 +106,9 @@ static int bfs_printf_ctime(FILE *file, const struct bfs_printf_directive *direc
 	return fprintf(file, directive->str, buf);
 }
 
-/** %A, %C, %T: strftime() */
+/** %A, %B/%W, %C, %T: strftime() */
 static int bfs_printf_strftime(FILE *file, const struct bfs_printf_directive *directive, const struct BFTW *ftwbuf) {
-	const struct timespec *ts = get_time_field(ftwbuf->statbuf, directive->stat_field);
+	const struct timespec *ts = bfs_stat_time(ftwbuf->statbuf, directive->stat_field);
 	if (!ts) {
 		return -1;
 	}
@@ -149,15 +126,6 @@ static int bfs_printf_strftime(FILE *file, const struct bfs_printf_directive *di
 	case '@':
 		ret = snprintf(buf, sizeof(buf), "%lld.%09ld0", (long long)ts->tv_sec, (long)ts->tv_nsec);
 		break;
-	case 'k':
-		ret = snprintf(buf, sizeof(buf), "%2d", tm.tm_hour);
-		break;
-	case 'l':
-		ret = snprintf(buf, sizeof(buf), "%2d", (tm.tm_hour + 11)%12 + 1);
-		break;
-	case 'S':
-		ret = snprintf(buf, sizeof(buf), "%.2d.%09ld0", tm.tm_sec, (long)ts->tv_nsec);
-		break;
 	case '+':
 		ret = snprintf(buf, sizeof(buf), "%4d-%.2d-%.2d+%.2d:%.2d:%.2d.%09ld0",
 		               1900 + tm.tm_year,
@@ -168,6 +136,25 @@ static int bfs_printf_strftime(FILE *file, const struct bfs_printf_directive *di
 		               tm.tm_sec,
 		               (long)ts->tv_nsec);
 		break;
+	case 'k':
+		ret = snprintf(buf, sizeof(buf), "%2d", tm.tm_hour);
+		break;
+	case 'l':
+		ret = snprintf(buf, sizeof(buf), "%2d", (tm.tm_hour + 11)%12 + 1);
+		break;
+	case 's':
+		ret = snprintf(buf, sizeof(buf), "%lld", (long long)ts->tv_sec);
+		break;
+	case 'S':
+		ret = snprintf(buf, sizeof(buf), "%.2d.%09ld0", tm.tm_sec, (long)ts->tv_nsec);
+		break;
+	case 'T':
+		ret = snprintf(buf, sizeof(buf), "%.2d:%.2d:%.2d.%09ld0",
+			       tm.tm_hour,
+			       tm.tm_min,
+			       tm.tm_sec,
+			       (long)ts->tv_nsec);
+		break;
 
 	// POSIX strftime() features
 	default:
@@ -184,7 +171,8 @@ static int bfs_printf_strftime(FILE *file, const struct bfs_printf_directive *di
 
 /** %b: blocks */
 static int bfs_printf_b(FILE *file, const struct bfs_printf_directive *directive, const struct BFTW *ftwbuf) {
-	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)ftwbuf->statbuf->blocks);
+	uintmax_t blocks = ((uintmax_t)ftwbuf->statbuf->blocks*BFS_STAT_BLKSIZE + 511)/512;
+	BFS_PRINTF_BUF(buf, "%ju", blocks);
 	return fprintf(file, directive->str, buf);
 }
 
@@ -266,7 +254,8 @@ static int bfs_printf_i(FILE *file, const struct bfs_printf_directive *directive
 
 /** %k: 1K blocks */
 static int bfs_printf_k(FILE *file, const struct bfs_printf_directive *directive, const struct BFTW *ftwbuf) {
-	BFS_PRINTF_BUF(buf, "%ju", (uintmax_t)(ftwbuf->statbuf->blocks + 1)/2);
+	uintmax_t blocks = ((uintmax_t)ftwbuf->statbuf->blocks*BFS_STAT_BLKSIZE + 1023)/1024;
+	BFS_PRINTF_BUF(buf, "%ju", blocks);
 	return fprintf(file, directive->str, buf);
 }
 
@@ -326,7 +315,13 @@ static int bfs_printf_s(FILE *file, const struct bfs_printf_directive *directive
 
 /** %S: sparseness */
 static int bfs_printf_S(FILE *file, const struct bfs_printf_directive *directive, const struct BFTW *ftwbuf) {
-	double sparsity = 512.0 * ftwbuf->statbuf->blocks / ftwbuf->statbuf->size;
+	double sparsity;
+	const struct bfs_stat *sb = ftwbuf->statbuf;
+	if (sb->size == 0 && sb->blocks == 0) {
+		sparsity = 1.0;
+	} else {
+		sparsity = (double)BFS_STAT_BLKSIZE*sb->blocks/sb->size;
+	}
 	return fprintf(file, directive->str, sparsity);
 }
 
@@ -377,6 +372,8 @@ static int bfs_printf_y(FILE *file, const struct bfs_printf_directive *directive
 
 /** %Y: target type */
 static int bfs_printf_Y(FILE *file, const struct bfs_printf_directive *directive, const struct BFTW *ftwbuf) {
+	int error = 0;
+
 	if (ftwbuf->typeflag != BFTW_LNK) {
 		return bfs_printf_y(file, directive, ftwbuf);
 	}
@@ -385,7 +382,7 @@ static int bfs_printf_Y(FILE *file, const struct bfs_printf_directive *directive
 
 	struct bfs_stat sb;
 	if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, 0, 0, &sb) == 0) {
-		type = bfs_printf_type(mode_to_typeflag(sb.mode));
+		type = bfs_printf_type(bftw_mode_typeflag(sb.mode));
 	} else {
 		switch (errno) {
 		case ELOOP:
@@ -395,10 +392,19 @@ static int bfs_printf_Y(FILE *file, const struct bfs_printf_directive *directive
 		case ENOTDIR:
 			type = "N";
 			break;
+		default:
+			type = "?";
+			error = errno;
+			break;
 		}
 	}
 
-	return fprintf(file, directive->str, type);
+	int ret = fprintf(file, directive->str, type);
+	if (error != 0) {
+		ret = -1;
+		errno = error;
+	}
+	return ret;
 }
 
 /**
@@ -414,7 +420,7 @@ static void free_directive(struct bfs_printf_directive *directive) {
 /**
  * Create a new printf directive.
  */
-static struct bfs_printf_directive *new_directive() {
+static struct bfs_printf_directive *new_directive(void) {
 	struct bfs_printf_directive *directive = malloc(sizeof(*directive));
 	if (!directive) {
 		perror("malloc()");
@@ -471,8 +477,6 @@ static int append_literal(struct bfs_printf_directive ***tail, struct bfs_printf
 }
 
 struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline) {
-	CFILE *cerr = cmdline->cerr;
-
 	struct bfs_printf *command = malloc(sizeof(*command));
 	if (!command) {
 		perror("malloc()");
@@ -527,11 +531,11 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				goto done;
 
 			case '\0':
-				cfprintf(cerr, "%{er}error: '%s': Incomplete escape sequence '\\'.%{rs}\n", format);
+				bfs_error(cmdline, "'%s': Incomplete escape sequence '\\'.\n", format);
 				goto error;
 
 			default:
-				cfprintf(cerr, "%{er}error: '%s': Unrecognized escape sequence '\\%c'.%{rs}\n", format, c);
+				bfs_error(cmdline, "'%s': Unrecognized escape sequence '\\%c'.\n", format, c);
 				goto error;
 			}
 		} else if (c == '%') {
@@ -561,10 +565,11 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				case '0':
 				case '+':
 					must_be_numeric = true;
+					// Fallthrough
 				case ' ':
 				case '-':
 					if (strchr(directive->str, c)) {
-						cfprintf(cerr, "%{er}error: '%s': Duplicate flag '%c'.%{rs}\n", format, c);
+						bfs_error(cmdline, "'%s': Duplicate flag '%c'.\n", format, c);
 						goto directive_error;
 					}
 					if (dstrapp(&directive->str, c) != 0) {
@@ -627,7 +632,7 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				if (!cmdline->mtab) {
 					cmdline->mtab = parse_bfs_mtab();
 					if (!cmdline->mtab) {
-						cfprintf(cmdline->cerr, "%{er}error: Couldn't parse the mount table: %m%{rs}\n");
+						bfs_error(cmdline, "Couldn't parse the mount table: %m.\n");
 						goto directive_error;
 					}
 				}
@@ -716,47 +721,43 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 			case 'A':
 				directive->stat_field = BFS_STAT_ATIME;
 				goto directive_strftime;
+			case 'B':
+			case 'W':
+				directive->stat_field = BFS_STAT_BTIME;
+				goto directive_strftime;
 			case 'C':
 				directive->stat_field = BFS_STAT_CTIME;
 				goto directive_strftime;
 			case 'T':
 				directive->stat_field = BFS_STAT_MTIME;
 				goto directive_strftime;
-			case 'W':
-				directive->stat_field = BFS_STAT_BTIME;
-				goto directive_strftime;
 
 			directive_strftime:
 				directive->fn = bfs_printf_strftime;
 				command->needs_stat = true;
 				c = *++i;
 				if (!c) {
-					cfprintf(cerr, "%{er}error: '%s': Incomplete time specifier '%s%c'.%{rs}\n",
-					         format, directive->str, i[-1]);
+					bfs_error(cmdline, "'%s': Incomplete time specifier '%s%c'.\n", format, directive->str, i[-1]);
 					goto directive_error;
-				} else if (strchr("@HIklMprST+XZaAbBcdDhjmUwWxyY", c)) {
+				} else if (strchr("%+@aAbBcCdDeFgGhHIjklmMnprRsStTuUVwWxXyYzZ", c)) {
 					directive->c = c;
 				} else {
-					cfprintf(cerr, "%{er}error: '%s': Unrecognized time specifier '%%%c%c'.%{rs}\n",
-					         format, i[-1], c);
+					bfs_error(cmdline, "'%s': Unrecognized time specifier '%%%c%c'.\n", format, i[-1], c);
 					goto directive_error;
 				}
 				break;
 
 			case '\0':
-				cfprintf(cerr, "%{er}error: '%s': Incomplete format specifier '%s'.%{rs}\n",
-				         format, directive->str);
+				bfs_error(cmdline, "'%s': Incomplete format specifier '%s'.\n", format, directive->str);
 				goto directive_error;
 
 			default:
-				cfprintf(cerr, "%{er}error: '%s': Unrecognized format specifier '%%%c'.%{rs}\n",
-				         format, c);
+				bfs_error(cmdline, "'%s': Unrecognized format specifier '%%%c'.\n", format, c);
 				goto directive_error;
 			}
 
 			if (must_be_numeric && strcmp(specifier, "s") == 0) {
-				cfprintf(cerr, "%{er}error: '%s': Invalid flags '%s' for string format '%%%c'.%{rs}\n",
-				         format, directive->str + 1, c);
+				bfs_error(cmdline, "'%s': Invalid flags '%s' for string format '%%%c'.\n", format, directive->str + 1, c);
 				goto directive_error;
 			}
 
@@ -798,16 +799,16 @@ error:
 }
 
 int bfs_printf(FILE *file, const struct bfs_printf *command, const struct BFTW *ftwbuf) {
-	int ret = -1;
+	int ret = 0, error = 0;
 
 	for (struct bfs_printf_directive *directive = command->directives; directive; directive = directive->next) {
 		if (directive->fn(file, directive, ftwbuf) < 0) {
-			goto done;
+			ret = -1;
+			error = errno;
 		}
 	}
 
-	ret = 0;
-done:
+	errno = error;
 	return ret;
 }
 
