@@ -16,9 +16,11 @@
 
 #include "stat.h"
 #include "util.h"
+#include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <stdbool.h>
+#include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -32,7 +34,6 @@
 
 #if HAVE_STATX || defined(__NR_statx)
 #	define HAVE_BFS_STATX true
-#	include <sys/sysmacros.h>
 #endif
 
 #if __APPLE__
@@ -42,6 +43,42 @@
 #	define st_birthtim st_birthtimespec
 #endif
 
+const char *bfs_stat_field_name(enum bfs_stat_field field) {
+	switch (field) {
+	case BFS_STAT_DEV:
+		return "device number";
+	case BFS_STAT_INO:
+		return "inode nunmber";
+	case BFS_STAT_TYPE:
+		return "type";
+	case BFS_STAT_MODE:
+		return "mode";
+	case BFS_STAT_NLINK:
+		return "link count";
+	case BFS_STAT_GID:
+		return "group ID";
+	case BFS_STAT_UID:
+		return "user ID";
+	case BFS_STAT_SIZE:
+		return "size";
+	case BFS_STAT_BLOCKS:
+		return "block count";
+	case BFS_STAT_RDEV:
+		return "underlying device";
+	case BFS_STAT_ATIME:
+		return "access time";
+	case BFS_STAT_BTIME:
+		return "birth time";
+	case BFS_STAT_CTIME:
+		return "change time";
+	case BFS_STAT_MTIME:
+		return "modification time";
+	}
+
+	assert(false);
+	return "???";
+}
+
 /**
  * Check if we should retry a failed stat() due to a potentially broken link.
  */
@@ -82,6 +119,9 @@ static void bfs_stat_convert(const struct stat *statbuf, struct bfs_stat *buf) {
 	buf->blocks = statbuf->st_blocks;
 	buf->mask |= BFS_STAT_BLOCKS;
 
+	buf->rdev = statbuf->st_rdev;
+	buf->mask |= BFS_STAT_RDEV;
+
 	buf->atime = statbuf->st_atim;
 	buf->mask |= BFS_STAT_ATIME;
 
@@ -122,6 +162,12 @@ static int bfs_stat_impl(int at_fd, const char *at_path, int at_flags, enum bfs_
  * Wrapper for the statx() system call, which had no glibc wrapper prior to 2.28.
  */
 static int bfs_statx(int at_fd, const char *at_path, int at_flags, unsigned int mask, struct statx *buf) {
+	// -fsanitize=memory doesn't know about statx(), so tell it the memory
+	// got initialized
+#if BFS_HAS_FEATURE(memory_sanitizer, false)
+	memset(buf, 0, sizeof(*buf));
+#endif
+
 #if HAVE_STATX
 	return statx(at_fd, at_path, at_flags, mask, buf);
 #else
@@ -146,15 +192,16 @@ static int bfs_statx_impl(int at_fd, const char *at_path, int at_flags, enum bfs
 		return ret;
 	}
 
-	if ((xbuf.stx_mask & STATX_BASIC_STATS) != STATX_BASIC_STATS) {
-		// Callers shouldn't have to check anything except BFS_STAT_BTIME
-		errno = EINVAL;
+	// Callers shouldn't have to check anything except the times
+	const unsigned int guaranteed = STATX_BASIC_STATS ^ (STATX_ATIME | STATX_CTIME | STATX_MTIME);
+	if ((xbuf.stx_mask & guaranteed) != guaranteed) {
+		errno = ENOTSUP;
 		return -1;
 	}
 
 	buf->mask = 0;
 
-	buf->dev = makedev(xbuf.stx_dev_major, xbuf.stx_dev_minor);
+	buf->dev = bfs_makedev(xbuf.stx_dev_major, xbuf.stx_dev_minor);
 	buf->mask |= BFS_STAT_DEV;
 
 	if (xbuf.stx_mask & STATX_INO) {
@@ -195,6 +242,9 @@ static int bfs_statx_impl(int at_fd, const char *at_path, int at_flags, enum bfs
 		buf->mask |= BFS_STAT_BLOCKS;
 	}
 
+	buf->rdev = bfs_makedev(xbuf.stx_rdev_major, xbuf.stx_rdev_minor);
+	buf->mask |= BFS_STAT_RDEV;
+
 	if (xbuf.stx_mask & STATX_ATIME) {
 		buf->atime.tv_sec = xbuf.stx_atime.tv_sec;
 		buf->atime.tv_nsec = xbuf.stx_atime.tv_nsec;
@@ -262,3 +312,25 @@ int bfs_fstat(int fd, struct bfs_stat *buf) {
 	}
 	return ret;
 }
+
+const struct timespec *bfs_stat_time(const struct bfs_stat *buf, enum bfs_stat_field field) {
+	if (!(buf->mask & field)) {
+		errno = ENOTSUP;
+		return NULL;
+	}
+
+	switch (field) {
+	case BFS_STAT_ATIME:
+		return &buf->atime;
+	case BFS_STAT_BTIME:
+		return &buf->btime;
+	case BFS_STAT_CTIME:
+		return &buf->ctime;
+	case BFS_STAT_MTIME:
+		return &buf->mtime;
+	default:
+		assert(false);
+		errno = EINVAL;
+		return NULL;
+	}
+}
