@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2018 Tavian Barnes <tavianator@tavianator.com>             *
+ * Copyright (C) 2018-2019 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -14,12 +14,26 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           *
  ****************************************************************************/
 
+/**
+ * A facade over the stat() API that unifies some details that diverge between
+ * implementations, like the names of the timespec fields and the presence of
+ * file "birth" times.  On new enough Linux kernels, the facade is backed by
+ * statx() instead, and so it exposes a similar interface with a mask for which
+ * fields were successfully returned.
+ */
+
 #ifndef BFS_STAT_H
 #define BFS_STAT_H
 
+#include "util.h"
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <time.h>
 
+#if BFS_HAS_SYS_PARAM
+#	include <sys/param.h>
+#endif
+
 /**
  * bfs_stat field bitmask.
  */
@@ -33,12 +47,18 @@ enum bfs_stat_field {
 	BFS_STAT_UID    = 1 << 6,
 	BFS_STAT_SIZE   = 1 << 7,
 	BFS_STAT_BLOCKS = 1 << 8,
-	BFS_STAT_ATIME  = 1 << 9,
-	BFS_STAT_BTIME  = 1 << 10,
-	BFS_STAT_CTIME  = 1 << 11,
-	BFS_STAT_MTIME  = 1 << 12,
+	BFS_STAT_RDEV   = 1 << 9,
+	BFS_STAT_ATIME  = 1 << 10,
+	BFS_STAT_BTIME  = 1 << 11,
+	BFS_STAT_CTIME  = 1 << 12,
+	BFS_STAT_MTIME  = 1 << 13,
 };
 
+/**
+ * Get the human-readable name of a bfs_stat field.
+ */
+const char *bfs_stat_field_name(enum bfs_stat_field field);
+
 /**
  * bfs_stat() flags.
  */
@@ -47,6 +67,14 @@ enum bfs_stat_flag {
 	BFS_STAT_BROKEN_OK = 1 << 0,
 };
 
+#ifdef DEV_BSIZE
+#	define BFS_STAT_BLKSIZE DEV_BSIZE
+#elif defined(S_BLKSIZE)
+#	define BFS_STAT_BLKSIZE S_BLKSIZE
+#else
+#	define BFS_STAT_BLKSIZE 512
+#endif
+
 /**
  * Facade over struct stat.
  */
@@ -68,8 +96,10 @@ struct bfs_stat {
 	uid_t uid;
 	/** File size in bytes. */
 	off_t size;
-	/** Number of 512B blocks allocated. */
+	/** Number of disk blocks allocated (of size BFS_STAT_BLKSIZE). */
 	blkcnt_t blocks;
+	/** The device ID represented by this file. */
+	dev_t rdev;
 
 	/** Access time. */
 	struct timespec atime;
@@ -91,4 +121,9 @@ int bfs_stat(int at_fd, const char *at_path, int at_flags, enum bfs_stat_flag fl
  */
 int bfs_fstat(int fd, struct bfs_stat *buf);
 
+/**
+ * Get a particular time field from a bfs_stat() buffer.
+ */
+const struct timespec *bfs_stat_time(const struct bfs_stat *buf, enum bfs_stat_field field);
+
 #endif // BFS_STAT_H
