@@ -2,7 +2,7 @@
 
 ############################################################################
 # bfs                                                                      #
-# Copyright (C) 2015-2017 Tavian Barnes <tavianator@tavianator.com>        #
+# Copyright (C) 2015-2018 Tavian Barnes <tavianator@tavianator.com>        #
 #                                                                          #
 # Permission to use, copy, modify, and/or distribute this software for any #
 # purpose with or without fee is hereby granted.                           #
@@ -22,6 +22,16 @@ umask 022
 export LC_ALL=C
 export TZ=UTC
 
+function _realpath() {
+    (
+        cd "${1%/*}"
+        echo "$PWD/${1##*/}"
+    )
+}
+
+BFS="$(_realpath ./bfs)"
+TESTS="$(_realpath ./tests)"
+
 # The temporary directory that will hold our test data
 TMP="$(mktemp -d "${TMPDIR:-/tmp}"/bfs.XXXXXXXXXX)"
 chown "$(id -u):$(id -g)" "$TMP"
@@ -53,7 +63,9 @@ function installp() {
 
 # Like a mythical touch -p
 function touchp() {
-    installp -m644 /dev/null "$1"
+    for arg; do
+        installp -m644 /dev/null "$arg"
+    done
 }
 
 # Creates a simple file+directory structure for tests
@@ -73,6 +85,7 @@ make_basic "$TMP/basic"
 
 # Creates a file+directory structure with various permissions for tests
 function make_perms() {
+    installp -m000 /dev/null "$1/0"
     installp -m444 /dev/null "$1/r"
     installp -m222 /dev/null "$1/w"
     installp -m644 /dev/null "$1/rw"
@@ -89,12 +102,25 @@ function make_links() {
     ln "$1/file" "$1/hardlink"
     ln -s nowhere "$1/broken"
     ln -s symlink/file "$1/notdir"
+    mkdir -p "$1/deeply/nested/dir"
+    touchp "$1/deeply/nested/file"
+    ln -s file "$1/deeply/nested/link"
+    ln -s deeply/nested "$1/skip"
+}
+make_links "$TMP/links"
+
+# Creates a file+directory structure with symbolic link loops
+function make_loops() {
+    touchp "$1/file"
+    ln -s file "$1/symlink"
+    ln -s nowhere "$1/broken"
+    ln -s symlink/file "$1/notdir"
     ln -s loop "$1/loop"
     mkdir -p "$1/deeply/nested/dir"
     ln -s ../../deeply "$1/deeply/nested/loop"
     ln -s deeply/nested/loop/nested "$1/skip"
 }
-make_links "$TMP/links"
+make_loops "$TMP/loops"
 
 # Creates a file+directory structure with varying timestamps
 function make_times() {
@@ -138,76 +164,109 @@ function make_deep() {
             mkdir "$1/$i"
             cd "$1/$i"
 
-            # 17 * 256 > 16 * 256 == 4096 == PATH_MAX
-            for j in {1..17}; do
+            # 16 * 256 == 4096 == PATH_MAX
+            for j in {1..16}; do
                 mkdir "$name"
                 cd "$name" 2>/dev/null
             done
+
+            touch "$name"
         )
     done
 }
 make_deep "$TMP/deep"
 
+# Creates a directory structure with many different types, and therefore colors
+function make_rainbow() {
+    touchp "$1/file.txt"
+    touchp "$1/file.dat"
+    ln -s file.txt "$1/link.txt"
+    touchp "$1/mh1"
+    ln "$1/mh1" "$1/mh2"
+    mkfifo "$1/pipe"
+    # TODO: block
+    # TODO: chardev
+    ln -s nowhere "$1/broken"
+    "$TESTS/mksock" "$1/socket"
+    touchp "$1"/s{u,g,ug}id
+    chmod u+s "$1"/su{,g}id
+    chmod g+s "$1"/s{u,}gid
+    mkdir "$1/ow" "$1"/sticky{,_ow}
+    chmod o+w "$1"/*ow
+    chmod +t "$1"/sticky*
+    touchp "$1"/exec.sh
+    chmod +x "$1"/exec.sh
+}
+make_rainbow "$TMP/rainbow"
+
 # Creates a scratch directory that tests can modify
 function make_scratch() {
     mkdir -p "$1"
 }
 make_scratch "$TMP/scratch"
 
-function _realpath() {
-    (
-        cd "${1%/*}"
-        echo "$PWD/${1##*/}"
-    )
-}
-
-BFS="$(_realpath ./bfs)"
-TESTS="$(_realpath ./tests)"
-
 posix_tests=(
+    # General parsing
     test_basic
-    test_type_d
-    test_type_f
-    test_mindepth
-    test_maxdepth
-    test_depth
-    test_depth_slash
-    test_depth_mindepth_1
-    test_depth_mindepth_2
-    test_depth_maxdepth_1
-    test_depth_maxdepth_2
-    test_depth_error
-    test_name
-    test_name_root
-    test_name_root_depth
-    test_name_trailing_slash
-    test_path
-    test_newer
-    test_newer_link
-    test_links
-    test_links_plus
-    test_links_minus
+
+    test_parens
+    test_bang
+    test_implicit_and
+    test_a
+    test_o
+
+    test_weird_names
+
+    # Flags
+
     test_H
     test_H_slash
     test_H_broken
     test_H_notdir
-    test_H_newer
+    test_H_loops
+
     test_L
     test_L_broken
     test_L_notdir
+    test_L_loops
+
+    test_flag_weird_names
+    test_flag_comma
+
+    # Primaries
+
+    test_depth
+    test_depth_slash
+    test_depth_error
     test_L_depth
-    test_user_name
-    test_user_id
-    test_group_name
-    test_group_id
-    test_size
-    test_size_plus
-    test_size_bytes
+
     test_exec
     test_exec_plus
     test_exec_plus_status
     test_exec_plus_semicolon
-    test_flag_comma
+
+    test_group_name
+    test_group_id
+
+    test_links
+    test_links_plus
+    test_links_minus
+
+    test_name
+    test_name_root
+    test_name_root_depth
+    test_name_trailing_slash
+
+    test_newer
+    test_newer_link
+
+    test_ok_stdin
+    test_ok_plus_semicolon
+
+    test_path
+
+    test_perm_000
+    test_perm_000_minus
     test_perm_222
     test_perm_222_minus
     test_perm_644
@@ -216,64 +275,66 @@ posix_tests=(
     test_perm_symbolic_minus
     test_perm_leading_plus_symbolic_minus
     test_permcopy
-    test_ok_stdin
-    test_ok_plus_semicolon
-    test_parens
-    test_bang
-    test_implicit_and
-    test_a
-    test_o
+
+    test_prune
+    test_prune_or_print
+    test_not_prune
+
+    test_size
+    test_size_plus
+    test_size_bytes
+
+    test_type_d
+    test_type_f
+
+    test_user_name
+    test_user_id
+
+    # Closed file descriptors
+    test_closed_stdin
+    test_closed_stdout
+    test_closed_stderr
+
+    # PATH_MAX handling
     test_deep
+
+    # Optimizer tests
+    test_or_purity
     test_double_negation
     test_de_morgan_not
     test_de_morgan_and
     test_de_morgan_or
     test_data_flow_type
+    test_data_flow_and_swap
+    test_data_flow_or_swap
 )
 
 bsd_tests=(
+    # Flags
+
+    test_E
+
     test_P
     test_P_slash
+
     test_X
-    test_follow
-    test_samefile
-    test_samefile_symlink
-    test_H_samefile_symlink
-    test_L_samefile_symlink
-    test_samefile_broken
-    test_H_samefile_broken
-    test_L_samefile_broken
-    test_samefile_notdir
-    test_H_samefile_notdir
-    test_L_samefile_notdir
-    test_name_slash
-    test_name_slashes
-    test_iname
-    test_ipath
-    test_lname
-    test_ilname
-    test_L_lname
-    test_L_ilname
-    test_newerma
-    test_size_big
-    test_exec_substring
-    test_execdir_pwd
-    test_execdir_slash
-    test_execdir_slash_pwd
-    test_execdir_slashes
+
+    test_d_path
+
+    test_f
+
     test_double_dash
     test_flag_double_dash
-    test_ok_stdin
-    test_okdir_stdin
+
+    # Primaries
+
     test_delete
-    test_L_delete
-    test_rm
-    test_regex
-    test_iregex
-    test_regex_parens
-    test_E
-    test_d_path
-    test_f
+
+    test_depth_maxdepth_1
+    test_depth_maxdepth_2
+    test_depth_mindepth_1
+    test_depth_mindepth_2
+
     test_depth_n
     test_depth_n_plus
     test_depth_n_minus
@@ -281,13 +342,67 @@ bsd_tests=(
     test_depth_depth_n_plus
     test_depth_depth_n_minus
     test_depth_overflow
+    test_data_flow_depth
+
+    test_exec_substring
+
+    test_execdir_pwd
+    test_execdir_slash
+    test_execdir_slash_pwd
+    test_execdir_slashes
+
+    test_exit
+
+    test_follow
+
     test_gid_name
-    test_uid_name
+
+    test_ilname
+    test_L_ilname
+
+    test_iname
+
+    test_inum
+
+    test_ipath
+
+    test_iregex
+
+    test_lname
+    test_L_lname
+
+    test_maxdepth
+
+    test_mindepth
+
     test_mnewer
     test_H_mnewer
+
+    test_name_slash
+    test_name_slashes
+
+    test_H_newer
+
+    test_newerma
+
+    test_nogroup
+    test_nogroup_ulimit
+
+    test_nouser
+    test_nouser_ulimit
+
+    test_ok_stdin
+    test_ok_closed_stdin
+
+    test_okdir_stdin
+    test_okdir_closed_stdin
+
+    test_perm_000_plus
     test_perm_222_plus
     test_perm_644_plus
-    test_size_T
+
+    test_printx
+
     test_quit
     test_quit_child
     test_quit_depth
@@ -295,31 +410,12 @@ bsd_tests=(
     test_quit_after_print
     test_quit_before_print
     test_quit_implicit_print
-    test_inum
-    test_nogroup
-    test_nouser
-    test_exit
-    test_printx
-    test_data_flow_depth
-)
 
-gnu_tests=(
-    test_true
-    test_false
-    test_executable
-    test_readable
-    test_writable
-    test_empty
-    test_gid
-    test_gid_plus
-    test_gid_minus
-    test_uid
-    test_uid_plus
-    test_uid_minus
-    test_anewer
-    test_P
-    test_P_slash
-    test_follow
+    test_rm
+
+    test_regex
+    test_regex_parens
+
     test_samefile
     test_samefile_symlink
     test_H_samefile_symlink
@@ -330,23 +426,57 @@ gnu_tests=(
     test_samefile_notdir
     test_H_samefile_notdir
     test_L_samefile_notdir
-    test_xtype_l
-    test_xtype_f
-    test_L_xtype_l
-    test_L_xtype_f
-    test_name_slash
-    test_name_slashes
-    test_iname
-    test_ipath
-    test_lname
-    test_ilname
-    test_L_lname
-    test_L_ilname
+
+    test_size_T
+    test_size_big
+
+    test_uid_name
+)
+
+gnu_tests=(
+    # General parsing
+
+    test_not
+    test_and
+    test_or
+    test_comma
+    test_precedence
+
+    test_follow_comma
+
+    # Flags
+
+    test_P
+    test_P_slash
+
+    test_L_loops_continue
+
+    test_double_dash
+    test_flag_double_dash
+
+    # Primaries
+
+    test_anewer
+
+    test_path_d
+
     test_daystart
     test_daystart_twice
-    test_newerma
-    test_size_big
+
+    test_delete
+    test_L_delete
+
+    test_depth_mindepth_1
+    test_depth_mindepth_2
+    test_depth_maxdepth_1
+    test_depth_maxdepth_2
+
+    test_empty
+
+
+    test_exec_nothing
     test_exec_substring
+
     test_execdir
     test_execdir_substring
     test_execdir_plus_semicolon
@@ -354,36 +484,72 @@ gnu_tests=(
     test_execdir_slash
     test_execdir_slash_pwd
     test_execdir_slashes
-    test_okdir_plus_semicolon
-    test_weird_names
-    test_flag_weird_names
-    test_follow_comma
+
+    test_executable
+
+    test_false
+
+    test_follow
+
     test_fprint
     test_fprint_duplicate
-    test_double_dash
-    test_flag_double_dash
+    test_fprint_error
+
+    test_fstype
+
+    test_gid
+    test_gid_plus
+    test_gid_minus
+
     test_ignore_readdir_race
     test_ignore_readdir_race_root
     test_ignore_readdir_race_notdir
+
+    test_ilname
+    test_L_ilname
+
+    test_iname
+
+    test_inum
+
+    test_ipath
+
+    test_iregex
+
+    test_lname
+    test_L_lname
+
+    test_maxdepth
+
+    test_mindepth
+
+    test_name_slash
+    test_name_slashes
+
+    test_H_newer
+
+    test_newerma
+
+    test_nogroup
+    test_nogroup_ulimit
+
+    test_nouser
+    test_nouser_ulimit
+
+    test_ok_closed_stdin
+    test_ok_nothing
+
+    test_okdir_closed_stdin
+    test_okdir_plus_semicolon
+
+    test_perm_000_slash
     test_perm_222_slash
     test_perm_644_slash
     test_perm_symbolic_slash
     test_perm_leading_plus_symbolic_slash
-    test_delete
-    test_L_delete
-    test_regex
-    test_iregex
-    test_regex_parens
-    test_regextype_posix_basic
-    test_regextype_posix_extended
-    test_path_d
-    test_quit
-    test_quit_child
-    test_quit_depth
-    test_quit_depth_child
-    test_inum
-    test_nogroup
-    test_nouser
+
+    test_print_error
+
     test_printf
     test_printf_slash
     test_printf_slashes
@@ -395,44 +561,129 @@ gnu_tests=(
     test_printf_times
     test_printf_leak
     test_printf_nul
+    test_printf_Y_error
+
+    test_quit
+    test_quit_child
+    test_quit_depth
+    test_quit_depth_child
     test_quit_after_print
     test_quit_before_print
-    test_fstype
-    test_not
-    test_and
-    test_or
-    test_comma
-    test_precedence
+
+    test_readable
+
+    test_regex
+    test_regex_parens
+
+    test_regextype_posix_basic
+    test_regextype_posix_extended
+
+    test_samefile
+    test_samefile_symlink
+    test_H_samefile_symlink
+    test_L_samefile_symlink
+    test_samefile_broken
+    test_H_samefile_broken
+    test_L_samefile_broken
+    test_samefile_notdir
+    test_H_samefile_notdir
+    test_L_samefile_notdir
+
+    test_size_big
+
+    test_true
+
+    test_uid
+    test_uid_plus
+    test_uid_minus
+
+    test_writable
+
+    test_xtype_l
+    test_xtype_f
+    test_L_xtype_l
+    test_L_xtype_f
+
+    # Optimizer tests
     test_and_purity
-    test_or_purity
     test_not_reachability
     test_comma_reachability
-    test_print_error
-    test_fprint_error
 )
 
 bfs_tests=(
-    test_type_multi
-    test_xtype_multi
-    test_xtype_reorder
-    test_perm_symbolic_trailing_comma
-    test_perm_symbolic_double_comma
-    test_perm_symbolic_missing_action
-    test_perm_leading_plus_symbolic
-    test_execdir_plus
-    test_hidden
-    test_nohidden
-    test_printf_w
+    # General parsing
     test_path_flag_expr
     test_path_expr_flag
     test_flag_expr_path
     test_expr_flag_path
     test_expr_path_flag
-    test_colors
+
+    # Primaries
+
+    test_color
+    test_color_mh
+    test_color_mh0
+    test_color_or
+    test_color_mi
+    test_color_or_mi
+    test_color_or0_mi
+    test_color_or_mi0
+    test_color_ext
+    test_color_ext0
+    test_color_missing_colon
+
+    test_execdir_plus
+
+    test_hidden
+
+    test_nohidden
+
+    test_perm_symbolic_trailing_comma
+    test_perm_symbolic_double_comma
+    test_perm_symbolic_missing_action
+    test_perm_leading_plus_symbolic
+
+    test_printf_w
+
+    test_type_multi
+
+    test_xtype_multi
+    test_xtype_reorder
+
+    # PATH_MAX handling
     test_deep_strict
-    test_deep_stricter
 )
 
+function usage() {
+    local pad=$(printf "%*s" ${#0} "")
+    cat <<EOF
+Usage: $0 [--bfs=path/to/bfs] [--posix|--bsd|--gnu|--all]
+       $pad [--noclean] [--update] [--verbose] [--help]
+       $pad [test_* [test_* ...]]
+
+  --bfs=path/to/bfs
+      Set the path to the bfs executable to test (default: ./bfs)
+
+  --posix|--bsd|--gnu|--all
+      Restrict the set of test cases to run
+
+  --noclean
+      Keep the test directories around after the run
+
+  --update
+      Update the expected outputs for the test cases
+
+  --verbose
+      Log the commands that get executed
+
+  --help
+      This message
+
+  test_*
+      Select individual test cases to run
+EOF
+}
+
 BSD=yes
 GNU=yes
 ALL=yes
@@ -477,12 +728,17 @@ for arg; do
         --verbose)
             VERBOSE=yes
             ;;
+        --help)
+            usage
+            exit 0
+            ;;
         test_*)
             EXPLICIT=yes
             enable_tests "$arg"
             ;;
         *)
-            echo "Unrecognized option '$arg'." >&2
+            printf "Unrecognized option '%s'.\n\n" "$arg" >&2
+            usage >&2
             exit 1
             ;;
     esac
@@ -544,6 +800,11 @@ function invoke_bfs() {
     $BFS "$@"
 }
 
+# Return value when bfs fails
+EX_BFS=10
+# Return value when a difference is detected
+EX_DIFF=20
+
 function bfs_diff() (
     bfs_verbose "$@"
 
@@ -559,9 +820,16 @@ function bfs_diff() (
     fi
 
     $BFS "$@" | bfs_sort >"$ACTUAL"
+    local STATUS="${PIPESTATUS[0]}"
 
     if [ ! "$UPDATE" ]; then
-        diff -u "$EXPECTED" "$ACTUAL"
+        diff -u "$EXPECTED" "$ACTUAL" || return $EX_DIFF
+    fi
+
+    if [ "$STATUS" -eq 0 ]; then
+        return 0
+    else
+        return $EX_BFS
     fi
 )
 
@@ -647,7 +915,7 @@ function test_depth_error() {
     chmod +r scratch/foo
     rm -rf scratch/*
 
-    return $ret
+    [ $ret -eq $EX_BFS ]
 }
 
 function test_name() {
@@ -778,8 +1046,12 @@ function test_H_newer() {
     bfs_diff -H times -newer times/l
 }
 
+function test_H_loops() {
+    bfs_diff -H loops/deeply/nested/loop
+}
+
 function test_L() {
-    bfs_diff -L links 2>/dev/null
+    bfs_diff -L links
 }
 
 function test_L_broken() {
@@ -790,16 +1062,29 @@ function test_L_notdir() {
     bfs_diff -H links/notdir
 }
 
+function test_L_loops() {
+    # POSIX says it's okay to either stop of keep going on seeing a filesystem
+    # loop, as long as a diagnostic is printed
+    local errors="$(invoke_bfs -L loops 2>&1 >/dev/null)"
+    [ -n "$errors" ]
+}
+
+function test_L_loops_continue() {
+    bfs_diff -L loops 2>/dev/null
+    [ $? -eq $EX_BFS ]
+}
+
 function test_X() {
     bfs_diff -X weirdnames 2>/dev/null
+    [ $? -eq $EX_BFS ]
 }
 
 function test_follow() {
-    bfs_diff links -follow 2>/dev/null
+    bfs_diff links -follow
 }
 
 function test_L_depth() {
-    bfs_diff -L links -depth 2>/dev/null
+    bfs_diff -L links -depth
 }
 
 function test_samefile() {
@@ -815,7 +1100,7 @@ function test_H_samefile_symlink() {
 }
 
 function test_L_samefile_symlink() {
-    bfs_diff -L links -samefile links/symlink 2>/dev/null
+    bfs_diff -L links -samefile links/symlink
 }
 
 function test_samefile_broken() {
@@ -827,7 +1112,7 @@ function test_H_samefile_broken() {
 }
 
 function test_L_samefile_broken() {
-    bfs_diff -L links -samefile links/broken 2>/dev/null
+    bfs_diff -L links -samefile links/broken
 }
 
 function test_samefile_notdir() {
@@ -839,27 +1124,27 @@ function test_H_samefile_notdir() {
 }
 
 function test_L_samefile_notdir() {
-    bfs_diff -L links -samefile links/notdir 2>/dev/null
+    bfs_diff -L links -samefile links/notdir
 }
 
 function test_xtype_l() {
-    bfs_diff links -xtype l 2>/dev/null
+    bfs_diff links -xtype l
 }
 
 function test_xtype_f() {
-    bfs_diff links -xtype f 2>/dev/null
+    bfs_diff links -xtype f
 }
 
 function test_L_xtype_l() {
-    bfs_diff -L links -xtype l 2>/dev/null
+    bfs_diff -L links -xtype l
 }
 
 function test_L_xtype_f() {
-    bfs_diff -L links -xtype f 2>/dev/null
+    bfs_diff -L links -xtype f
 }
 
 function test_xtype_multi() {
-    bfs_diff links -xtype f,d,c 2>/dev/null
+    bfs_diff links -xtype f,d,c
 }
 
 function test_xtype_reorder() {
@@ -938,7 +1223,12 @@ function test_size_big() {
 }
 
 function test_exec() {
-    bfs_diff basic -exec echo '{}' ';'
+    bfs_diff basic -exec echo '{}' \;
+}
+
+function test_exec_nothing() {
+    # Regression test: don't segfault on missing command
+    ! invoke_bfs basic -exec \; 2>/dev/null
 }
 
 function test_exec_plus() {
@@ -1001,12 +1291,12 @@ function test_execdir_slashes() {
 
 function test_weird_names() {
     cd weirdnames
-    bfs_diff '-' '(-' '!-' ',' ')' './(' './!' \( \! -print , -print \)
+    bfs_diff '-' '(-' '!-' ',' ')' './(' './!' \( \! -print -o -print \)
 }
 
 function test_flag_weird_names() {
     cd weirdnames
-    bfs_diff -L '-' '(-' '!-' ',' ')' './(' './!' \( \! -print , -print \)
+    bfs_diff -L '-' '(-' '!-' ',' ')' './(' './!' \( \! -print -o -print \)
 }
 
 function test_flag_comma() {
@@ -1078,6 +1368,22 @@ function test_ignore_readdir_race_notdir() {
     invoke_bfs scratch -mindepth 1 -ignore_readdir_race -execdir rm -r '{}' \; -execdir touch '{}' \;
 }
 
+function test_perm_000() {
+    bfs_diff perms -perm 000
+}
+
+function test_perm_000_minus() {
+    bfs_diff perms -perm -000
+}
+
+function test_perm_000_slash() {
+    bfs_diff perms -perm /000
+}
+
+function test_perm_000_plus() {
+    bfs_diff perms -perm +000
+}
+
 function test_perm_222() {
     bfs_diff perms -perm 222
 }
@@ -1150,15 +1456,32 @@ function test_permcopy() {
     bfs_diff perms -perm u+rw,g+u-w,o=g
 }
 
+function test_prune() {
+    bfs_diff basic -name foo -prune
+}
+
+function test_prune_or_print() {
+    bfs_diff basic -name foo -prune -o -print
+}
+
+function test_not_prune() {
+    bfs_diff basic \! \( -name foo -prune \)
+}
+
+function test_ok_nothing() {
+    # Regression test: don't segfault on missing command
+    ! invoke_bfs basic -ok \; 2>/dev/null
+}
+
 function test_ok_stdin() {
     # -ok should *not* close stdin
     # See https://savannah.gnu.org/bugs/?24561
-    yes | bfs_diff basic -ok bash -c "printf '%s? ' {} && head -n1" \; 2>/dev/null
+    yes | bfs_diff basic -ok bash -c 'printf "%s? " "$1" && head -n1' bash '{}' \; 2>/dev/null
 }
 
 function test_okdir_stdin() {
     # -okdir should *not* close stdin
-    yes | bfs_diff basic -okdir bash -c "printf '%s? ' {} && head -n1" \; 2>/dev/null
+    yes | bfs_diff basic -okdir bash -c 'printf "%s? " "$1" && head -n1' bash '{}' \; 2>/dev/null
 }
 
 function test_ok_plus_semicolon() {
@@ -1174,7 +1497,7 @@ function test_delete() {
     touchp scratch/foo/bar/baz
 
     # Don't try to delete '.'
-    (cd scratch && invoke_bfs -delete)
+    (cd scratch && invoke_bfs . -delete)
 
     bfs_diff scratch
 }
@@ -1195,7 +1518,7 @@ function test_rm() {
     rm -rf scratch/*
     touchp scratch/foo/bar/baz
 
-    (cd scratch && invoke_bfs -rm)
+    (cd scratch && invoke_bfs . -rm)
 
     bfs_diff scratch
 }
@@ -1334,10 +1657,22 @@ function test_nogroup() {
     bfs_diff basic -nogroup
 }
 
+function test_nogroup_ulimit() {
+    closefrom 4
+    ulimit -n 16
+    bfs_diff deep -nogroup
+}
+
 function test_nouser() {
     bfs_diff basic -nouser
 }
 
+function test_nouser_ulimit() {
+    closefrom 4
+    ulimit -n 16
+    bfs_diff deep -nouser
+}
+
 function test_printf() {
     bfs_diff basic -printf '%%p(%p) %%d(%d) %%f(%f) %%h(%h) %%H(%H) %%P(%P) %%m(%m) %%M(%M) %%y(%y)\n'
 }
@@ -1363,7 +1698,7 @@ function test_printf_flags() {
 }
 
 function test_printf_types() {
-    bfs_diff links -printf '(%p) (%l) %y %Y\n'
+    bfs_diff loops -printf '(%p) (%l) %y %Y\n'
 }
 
 function test_printf_escapes() {
@@ -1396,8 +1731,23 @@ function test_printf_nul() {
 }
 
 function test_printf_w() {
-    # Birth times may not be supported, so just check that %w/%W can be parsed
-    bfs_diff times -false -printf '%w %WY\n'
+    # Birth times may not be supported, so just check that %w/%W/%B can be parsed
+    bfs_diff times -false -printf '%w %WY %BY\n'
+}
+
+function test_printf_Y_error() {
+    rm -rf scratch/*
+    mkdir scratch/foo
+    chmod -x scratch/foo
+    ln -s foo/bar scratch/bar
+
+    bfs_diff scratch -printf '(%p) (%l) %y %Y\n' 2>/dev/null
+    local ret=$?
+
+    chmod +x scratch/foo
+    rm -rf scratch/*
+
+    [ $ret -eq $EX_BFS ]
 }
 
 function test_fstype() {
@@ -1465,15 +1815,55 @@ function test_precedence() {
     bfs_diff basic \( -name foo -type d -o -name bar -a -type f \) -print , \! -empty -type f -print
 }
 
-function test_colors() {
-    LS_COLORS= bfs_diff links -color
+function test_color() {
+    LS_COLORS= bfs_diff rainbow -color
+}
+
+function test_color_mh() {
+    LS_COLORS="mh=01:" bfs_diff rainbow -color
+}
+
+function test_color_mh0() {
+    LS_COLORS="mh=00:" bfs_diff rainbow -color
+}
+
+function test_color_or() {
+    LS_COLORS="or=01:" bfs_diff rainbow -color
+}
+
+function test_color_mi() {
+    LS_COLORS="mi=01:" bfs_diff rainbow -color
+}
+
+function test_color_or_mi() {
+    LS_COLORS="or=01;31:mi=01;33:" bfs_diff rainbow -color
+}
+
+function test_color_or0_mi() {
+    LS_COLORS="or=00:mi=01;33:" bfs_diff rainbow -color
+}
+
+function test_color_or_mi0() {
+    LS_COLORS="or=01;31:mi=00:" bfs_diff rainbow -color
+}
+
+function test_color_ext() {
+    LS_COLORS="*.txt=01:" bfs_diff rainbow -color
+}
+
+function test_color_ext0() {
+    LS_COLORS="*.txt=00:" bfs_diff rainbow -color
+}
+
+function test_color_missing_colon() {
+    LS_COLORS="*.txt=01" bfs_diff rainbow -color
 }
 
 function test_deep() {
     closefrom 4
 
     ulimit -n 16
-    bfs_diff deep -mindepth 18 -exec /bin/bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' /bin/bash '{}' \;
+    bfs_diff deep -type f -exec bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' bash '{}' \;
 }
 
 function test_deep_strict() {
@@ -1481,15 +1871,7 @@ function test_deep_strict() {
 
     # Not even enough fds to keep the root open
     ulimit -n 7
-    bfs_diff deep -mindepth 18 -exec /bin/bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' /bin/bash '{}' \;
-}
-
-function test_deep_stricter() {
-    closefrom 4
-
-    # So few fds that pipe() fails in the -exec implementation
-    ulimit -n 6
-    bfs_diff deep -mindepth 18 -exec /bin/bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' /bin/bash '{}' \;
+    bfs_diff deep -type f -exec bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' bash '{}' \;
 }
 
 function test_exit() {
@@ -1503,7 +1885,7 @@ function test_exit() {
         return 1
     fi
 
-    bfs_diff basic -name bar -exit -o -print
+    bfs_diff basic/g -print -name g -exit
 }
 
 function test_printx() {
@@ -1552,6 +1934,14 @@ function test_data_flow_type() {
     bfs_diff basic \! \( -type f -o \! -type f \)
 }
 
+function test_data_flow_and_swap() {
+    bfs_diff basic \! -type f -a -type d
+}
+
+function test_data_flow_or_swap() {
+    bfs_diff basic \! \( -type f -o \! -type d \)
+}
+
 function test_print_error() {
     if [ -e /dev/full ]; then
         ! invoke_bfs basic -maxdepth 0 >/dev/full 2>/dev/null
@@ -1564,6 +1954,26 @@ function test_fprint_error() {
     fi
 }
 
+function test_closed_stdin() {
+    bfs_diff basic <&-
+}
+
+function test_ok_closed_stdin() {
+    bfs_diff basic -ok echo \; <&- 2>/dev/null
+}
+
+function test_okdir_closed_stdin() {
+    bfs_diff basic -okdir echo {} \; <&- 2>/dev/null
+}
+
+function test_closed_stdout() {
+    ! invoke_bfs basic >&- 2>/dev/null
+}
+
+function test_closed_stderr() {
+    ! invoke_bfs basic >&- 2>&-
+}
+
 if [ -t 1 -a ! "$VERBOSE" ]; then
     in_place=yes
 fi
